﻿using Nucleus.Gaming;
using SplitTool.Controls;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SplitTool
{
    public partial class MainForm : Form
    {
        protected GameManager gameManager;
        protected GameInfo info;

        protected List<ScreenControl> loaded_screens;
        protected ContextMenuStrip screenStrip;
        protected List<PlayerInfo> players;

        public MainForm()
        {
            AppDomain.CurrentDomain.UnhandledException += CurrentDomain_UnhandledException;

            InitializeComponent();

            gameManager = new GameManager();
            loaded_screens = new List<ScreenControl>();
            players = new List<PlayerInfo>();

            gamesCombo.DataSource = gameManager.Games;

            MakeContextStrip();
            MakeMonitors();

            numPlayers_ValueChanged(null, EventArgs.Empty);
        }

        void CurrentDomain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
        {
            MessageBox.Show(e.ExceptionObject.ToString());

            try
            {
                gameManager.End();
                if (handler != null)
                {
                    handler.End();
                }
            }
            catch { }

#if DEBUG

            try
            {
                if (e.ExceptionObject is Exception)
                {
                    SaveException((Exception)e.ExceptionObject);
                }
            }
            catch { }
#endif
        }

        private void SaveException(Exception e)
        {
            // Save Log of the Call Stack
            string log = GetLog(e);
            Assembly myAss = Assembly.GetEntryAssembly();
            string dir = Path.GetDirectoryName(myAss.Location);
            DateTime time = DateTime.Now;
            string logLoc = Path.Combine(dir, "Log " + time.Hour + "-" + time.Minute + "-" + time.Second) + ".log";

            using (FileStream stream = new FileStream(logLoc, FileMode.Create))
            {
                StreamWriter writer = new StreamWriter(stream);
                writer.Write(log);
            }
        }

        public string GetLog(Exception ex)
        {
            string log = "";
            Assembly myAss = Assembly.GetEntryAssembly();

            DateTime now = DateTime.Now;

            log += "// Nucleus Exception log";
            log += Environment.NewLine;
            log += now.ToShortDateString();
            log += Environment.NewLine;
            log += now.ToShortTimeString();
            log += Environment.NewLine;

            log += myAss.FullName;
            log += Environment.NewLine;

            log += ex.ToString();
            log += Environment.NewLine;
            log += ex.GetType().FullName;
            log += Environment.NewLine;
#if WINRT
            log += ex.GetType().GetTypeInfo().Assembly.FullName;
#else
            log += ex.GetType().Assembly.FullName;
#endif

            return log;
        }

        protected void AdjustContextStrip()
        {
            for (int i = 0; i < 4; i++)
            {
                screenStrip.Items[i].Enabled = PlayerCount > i;
            }
        }

        protected void MakeContextStrip()
        {
            screenStrip = new ContextMenuStrip();
            for (int i = 0; i < 4; i++)
            {
                int player = i + 1;

                ToolStripMenuItem item = new ToolStripMenuItem("Player " + player);
                var fullscreen = item.DropDownItems.Add("Fullscreen");
                fullscreen.Click += ScreenClick;

                ToolStripMenuItem horizontal = new ToolStripMenuItem("Horizontal");
                var halfTop = horizontal.DropDownItems.Add("Half Top");
                var halfBottom = horizontal.DropDownItems.Add("Half Bottom");
                halfTop.Click += ScreenClick;
                halfBottom.Click += ScreenClick;
                item.DropDownItems.Add(horizontal);

                ToolStripMenuItem vertical = new ToolStripMenuItem("Vertical");
                var halfLeft = vertical.DropDownItems.Add("Half Left");
                var halfRight = vertical.DropDownItems.Add("Half Right");
                halfLeft.Click += ScreenClick;
                halfRight.Click += ScreenClick;
                item.DropDownItems.Add(vertical);

                ToolStripMenuItem fourPlayers = new ToolStripMenuItem("3+ Players on 1 Screen");
                var topLeft = fourPlayers.DropDownItems.Add("Top Left");
                var topRight = fourPlayers.DropDownItems.Add("Top Right");
                var bottomLeft = fourPlayers.DropDownItems.Add("Bottom Left");
                var bottomRight = fourPlayers.DropDownItems.Add("Bottom Right");
                topLeft.Click += ScreenClick;
                topRight.Click += ScreenClick;
                bottomLeft.Click += ScreenClick;
                bottomRight.Click += ScreenClick;
                item.DropDownItems.Add(fourPlayers);

                ScreenHolder holda = new ScreenHolder(ScreenType.Fullscreen, player);
                fullscreen.Tag = holda;

                holda = new ScreenHolder(ScreenType.HorizontalTop, player);
                halfTop.Tag = holda;
                holda = new ScreenHolder(ScreenType.HorizontalBottom, player);
                halfBottom.Tag = holda;
                holda = new ScreenHolder(ScreenType.VerticalLeft, player);
                halfLeft.Tag = holda;
                holda = new ScreenHolder(ScreenType.VerticalRight, player);
                halfRight.Tag = holda;

                holda = new ScreenHolder(ScreenType.TopLeft, player);
                topLeft.Tag = holda;
                holda = new ScreenHolder(ScreenType.TopRight, player);
                topRight.Tag = holda;
                holda = new ScreenHolder(ScreenType.BottomLeft, player);
                bottomLeft.Tag = holda;
                holda = new ScreenHolder(ScreenType.BottomRight, player);
                bottomRight.Tag = holda;

                screenStrip.Items.Add(item);
            }
        }

        protected void ScreenClick(object sender, EventArgs e)
        {
            if (hover == null)
            {
                return;
            }
            ToolStripDropDownItem item = (ToolStripDropDownItem)sender;
            ScreenHolder holda = (ScreenHolder)item.Tag;

            for (int i = 0; i < loaded_screens.Count; i++)
            {
                var mscreen = loaded_screens[i];
                for (int j = 0; j < mscreen.Controls.Count; j++)
                {
                    Control con = mscreen.Controls[j];

                    if (con is PlayerControl)
                    {
                        PlayerControl player = (PlayerControl)con;
                        if (player.Player.Player == holda.Player)
                        {
                            player.Dispose();
                            hover.Controls.Remove(player);
                            players.Remove(player.Player);
                        }
                    }
                }
            }

            for (int i = 0; i < hover.Controls.Count; i++)
            {
                Control con = hover.Controls[i];

                if (con is PlayerControl)
                {
                    PlayerControl player = (PlayerControl)con;
                    PlayerInfo otha = player.Player;

                    // if one coincides with the other, remove the other one
                    if (otha.ScreenType == ScreenType.Fullscreen ||
                        otha.ScreenType == holda.Position ||
                        holda.Position == ScreenType.HorizontalTop &&
                        (otha.ScreenType == ScreenType.TopLeft || otha.ScreenType == ScreenType.TopRight || otha.ScreenType == ScreenType.VerticalLeft || otha.ScreenType == ScreenType.VerticalRight) ||
                        holda.Position  == ScreenType.HorizontalBottom &&
                        (otha.ScreenType == ScreenType.BottomLeft || otha.ScreenType == ScreenType.BottomRight || otha.ScreenType == ScreenType.VerticalLeft || otha.ScreenType == ScreenType.VerticalRight) ||
                        holda.Position == ScreenType.VerticalLeft &&
                        (otha.ScreenType == ScreenType.TopLeft || otha.ScreenType == ScreenType.BottomLeft || otha.ScreenType == ScreenType.HorizontalTop || otha.ScreenType == ScreenType.HorizontalBottom) ||
                        holda.Position == ScreenType.VerticalRight &&
                        (otha.ScreenType == ScreenType.TopRight || otha.ScreenType == ScreenType.BottomRight || otha.ScreenType == ScreenType.HorizontalTop || otha.ScreenType == ScreenType.HorizontalBottom))
                    {
                        // Love boolean Math XD
                        player.Dispose();
                        hover.Controls.Remove(player);
                        players.Remove(otha);
                    }
                }
            }

            PlayerInfo playa = new PlayerInfo();
            playa.ScreenIndex = this.loaded_screens.IndexOf(hover);
            playa.ScreenType = holda.Position;
            playa.Player = holda.Player;

            PlayerControl theplayer = new PlayerControl(playa);
            theplayer.ChangeName("Player " + holda.Player);
            hover.Controls.Add(theplayer);
            theplayer.BringToFront();
            theplayer.Accomodate();
            players.Add(playa);
        }

        protected virtual void MakeMonitors()
        {
            this.panelScreens.Controls.Clear();

            Screen[] screens = Screen.AllScreens;
            int minX = 0;
            int minY = 0;
            int maxWidth = 0;
            int maxHeight = 0;
            for (int i = 0; i < screens.Length; i++)
            {
                var screen = screens[i];
                var bounds = screen.Bounds;

                minX = Math.Min(minX, bounds.X);
                minY = Math.Min(minY, bounds.Y);
                maxWidth += bounds.Width;
                maxHeight += bounds.Height;
            }

            int width = panelScreens.Width;
            int height = panelScreens.Height;
            int maxSize = Math.Max(maxWidth, maxHeight);
            int size = Math.Max(width, height);
            float factor = maxSize / (float)size;

            for (int i = 0; i < screens.Length; i++)
            {
                var screen = screens[i];
                var bounds = screen.Bounds;

                int swidth = bounds.Width;
                int sheight = bounds.Height;
                int x = bounds.X;
                int y = bounds.Y;

                swidth = (int)(swidth / factor);
                sheight = (int)(sheight / factor);
                x = (int)(x / factor);
                y = (int)(y / factor);

                ScreenControl scr = new ScreenControl();
                scr.MouseEnter += scr_MouseEnter;
                scr.ContextMenuStrip = screenStrip;
                scr.Size = new Size(swidth, sheight);
                scr.Location = new Point(x, y);
                scr.ChangeName(screen.DeviceName.Remove(0, screen.DeviceName.Length - 1));
                scr.Tag = screen;
                this.panelScreens.Controls.Add(scr);
                loaded_screens.Add(scr);
            }
        }

        protected ScreenControl hover;
        void scr_MouseEnter(object sender, EventArgs e)
        {
            hover = (ScreenControl)sender;
        }

        protected int PlayerCount = 1;
        private void timer1_Tick(object sender, EventArgs e)
        {
            if (loaded_screens != null)
            {
                if (Screen.AllScreens.Length != loaded_screens.Count)
                {
                    MakeMonitors();
                }
            }
        }

        private void btn_Browse_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog dialog = new OpenFileDialog())
            {
                dialog.Filter = "Game Starter|*.exe";
                if (dialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = dialog.FileName;
                    this.txt_GameFile.Text = fileName;
                }
            }
        }

        private void gamesCombo_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (gamesCombo.SelectedItem != null)
            {
                info = (GameInfo)gamesCombo.SelectedItem;
                numPlayers.Maximum = info.MaxPlayers;
                this.label_GameName.Text = info.GameName;

                controlListBox1.Controls.Clear();

                var options = info.Options;
                foreach (var opt in options.Values)
                {
                    CoolListControl cool = new CoolListControl();
                    cool.Text = opt.Name;
                    cool.Description = opt.Description;
                    cool.Width = this.controlListBox1.Width;

                    this.controlListBox1.Controls.Add(cool);

                    // Check the value type and add a control for it
                    if (opt.Value is bool)
                    {
                        SizeableCheckbox box = new SizeableCheckbox();
                        int border = 10;

                        box.Checked = (bool)opt.Value;
                        box.Width = 40;
                        box.Height = 40;
                        box.Left = cool.Width - box.Width - border;
                        box.Top = (cool.Height / 2) - (box.Height / 2);
                        cool.AddControl(box, false);

                        box.Tag = opt;
                        box.CheckedChanged += box_CheckedChanged;
                    }
                    else if (opt.Value is int)
                    {
                        NumericUpDown num = new NumericUpDown();
                        int border = 10;

                        num.Value = (int)opt.Value;

                        num.Width = 150;
                        num.Height = 40;
                        num.Left = cool.Width - num.Width - border;
                        num.Top = (cool.Height / 2) - (num.Height / 2);
                        cool.AddControl(num, false);

                        num.Tag = opt;
                        num.ValueChanged += num_ValueChanged;
                    }
                }
            }
        }

        void num_ValueChanged(object sender, EventArgs e)
        {
            NumericUpDown check = (NumericUpDown)sender;
            GameOption option = (GameOption)check.Tag;
            option.Value = (int)check.Value;
        }

        void box_CheckedChanged(object sender, EventArgs e)
        {
            SizeableCheckbox check = (SizeableCheckbox)sender;
            GameOption option = (GameOption)check.Tag;
            option.Value = check.Checked;
        }

        public event Action<Graphics> NewPaint;

        private void graphicalOverlay1_Paint(object sender, PaintEventArgs e)
        {
            if (NewPaint != null)
            {
                NewPaint(e.Graphics);
            }
        }

        private void numPlayers_ValueChanged(object sender, EventArgs e)
        {
            PlayerCount = (int)numPlayers.Value;
            AdjustContextStrip();
        }

        IGameHandler handler;
        private void btn_Play_Click(object sender, EventArgs e)
        {
            if (handler != null)
            {
                return;
            }
            string starter = this.txt_GameFile.Text;
            if (string.IsNullOrEmpty(starter))
            {
                MessageBox.Show("Select your game's starter file first!");
                return;
            }

            if (this.info == null)
            {
                return;
            }

            Type t = info.HandlerType;
            handler = (IGameHandler)Activator.CreateInstance(t);

            Rectangle screenRectangle = RectangleToScreen(this.ClientRectangle);
            int titleHeight = screenRectangle.Top - this.Top;
            bool worked = handler.Initialize(starter, this.players, info.Options, null, titleHeight);

            if (worked)
            {
                string error = handler.Play();

                if (string.IsNullOrEmpty(error))
                {
                    btn_Play.Enabled = false;
                    btn_End.Enabled = true;

                    gameManager.Play(handler);
                }
                else
                {
                    handler = null;
                    MessageBox.Show(error);
                }
            }
            else
            {
                handler = null;
            }
        }

        private void timer2_Tick(object sender, EventArgs e)
        {
            if (handler != null)
            {
                handler.Update(timer2.Interval);
            }
        }

        private void btn_End_Click(object sender, EventArgs e)
        {
            if (handler != null)
            {
                this.handler.End();
                this.handler = null;
            }
            gameManager.End();
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (handler != null)
            {
                handler.End();
                this.handler = null;
            }
            gameManager.End();
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (handler != null)
            {
                handler.End();
                this.handler = null;
            }
            gameManager.End();
        }
    }
}
